import itertools
from collections import deque
from typing import Callable, Hashable, Iterable, Optional, TypeVar

import toolz
import toolz.curried

T = TypeVar("T")


def closest_common(
    things: Iterable[Iterable[T]],
    get_key: Callable[[T], Hashable] = toolz.identity,
    validate_common_root: bool = False,
) -> T:
    """Returns the closest common item between a set of iterables."""
    if not things:
        raise ValueError("No things given.")

    things = iter(things)

    # make a yardstick of the first iterable
    # this is a dict with the keys being the __key() of the items
    # and the values being the index of the item in the iterable
    index_and_item = itertools.tee(enumerate(next(things)))
    key_to_index_map = dict((get_key(item), i) for i, item in index_and_item[0])
    index_to_item_map = dict((i, item) for i, item in index_and_item[1])

    # set the index of the common item
    # this starts at 0 because if there's no other item we
    # just want to return the first item
    common_i = 0
    max_i = len(key_to_index_map) - 1

    def shortcut_if_common_is_already_root():
        if not validate_common_root:
            if common_i == max_i:
                # return early in the case we're already at the root
                return index_to_item_map[common_i]

    shortcut_if_common_is_already_root()

    for thing in things:
        for item in thing:
            key = get_key(item)
            if key in key_to_index_map:
                # if the item is in the yardstick, then we've found a common item
                # if the new common item is further than the old common item, we update it
                item_common_i = key_to_index_map[key]
                if item_common_i > common_i:
                    common_i = item_common_i
                    shortcut_if_common_is_already_root()
                break
        else:
            # if we didn't find a common item, then we raise an error
            raise ValueError("No common item found.")

    return index_to_item_map[common_i]


# NOTE:
# 1. For the dfs and bfs functions, I'm not convinced we want to yield the start first
#    I think it might be a bit annoying in some cases...
# 2. The get_children function is first for both functions so we can curry them, and so their
#    signatures are the same as map and filter


def dfs_postorder(
    get_children: Callable[[T], Iterable[T]],
    start: T,
) -> Iterable[T]:
    """Depth-first search, yielding the leaves first."""

    for child in get_children(start):
        yield from dfs_postorder(get_children, child)

    yield start


def bfs(
    get_children: Callable[[T], Iterable[T]],
    start: T,
) -> Iterable[T]:
    """Breadth-first search, first yielding the starting item."""
    yield start

    queue = deque(get_children(start))

    while queue:
        start = queue.popleft()
        yield start

        for child in get_children(start):
            queue.append(child)


def recurse(
    get_next: Callable[[T], Optional[T]],
    start: T,
) -> Iterable[T]:
    """Recursively yield items, optionally including the starting item and its children."""
    yield start
    next_item = get_next(start)
    if next_item is not None:
        yield from recurse(get_next, next_item)



import hashlib
import uuid
from pathlib import Path
from typing import Optional

from jinja2 import Environment, FileSystemLoader, StrictUndefined
from toolz import groupby

from atopile import components, errors, nets
from atopile.address import AddrStr, get_name, get_relative_addr_str
from atopile.instance_methods import (
    all_descendants,
    get_children,
    get_next_super,
    get_parent,
    match_components,
    match_pins,
)
from atopile.kicad6_datamodel import (
    KicadComponent,
    KicadLibpart,
    KicadNet,
    KicadNetlist,
    KicadNode,
    KicadPin,
    KicadSheetpath,
)


_get_mpn = errors.downgrade(
    components.get_mpn, (components.MissingData, components.NoMatchingComponent)
)
_get_value = errors.downgrade(
    components.get_user_facing_value,
    (components.MissingData, components.NoMatchingComponent),
    default="?",
)


def generate_uid_from_path(path: str) -> str:
    """Spits out a uuid in hex from a string"""
    path_as_bytes = path.encode("utf-8")
    hashed_path = hashlib.blake2b(path_as_bytes, digest_size=16).digest()
    return str(uuid.UUID(bytes=hashed_path))


class NetlistBuilder:
    """He builds netlists."""

    def __init__(self):
        """TODO:"""
        self.netlist: Optional[KicadNetlist] = None
        self._libparts: dict[tuple, KicadLibpart] = {}
        self._components: dict[tuple, KicadComponent] = {}
        self._nets: dict[str, KicadNet] = {}

    def make_kicad_pin(self, pin_addr) -> KicadPin:
        """Make a KiCAD pin object from a representative instance object."""
        return KicadPin(name=get_name(pin_addr), type="stereo")

    def make_libpart(self, comp_addr: AddrStr) -> KicadLibpart:
        """Make a KiCAD libpart object from a representative instance object."""
        model_pins = filter(match_pins, get_children(comp_addr))
        pins = [self.make_kicad_pin(pin) for pin in model_pins]
        # def _get_origin_of_instance(instance: Instance) -> Object:
        #     return instance.origin

        super_abs_addr = get_next_super(comp_addr).obj_def.address
        super_addr = get_relative_addr_str(super_abs_addr)
        constructed_libpart = KicadLibpart(
            part=_get_mpn(comp_addr),
            description=super_addr,
            fields=[],
            pins=pins,
            # TODO: something better for these:
            lib="lib",
            docs="~",
            footprints=["*"],
        )
        return constructed_libpart

    def make_node(self, node_addr: AddrStr) -> KicadNode:
        """Make a KiCAD node object from a representative instance object."""
        parent = get_parent(node_addr)
        node = KicadNode(
            pin=get_name(node_addr),  # eg. 1
            ref=components.get_designator(parent),  # eg. R1
            pintype="stereo",
        )
        return node

    def make_net(self, code, net_name, net_list) -> KicadNet:
        """Make a KiCAD net object from a representative instance object."""
        net = KicadNet(
            code=code,
            name=net_name,
            nodes=[self.make_node(pin) for pin in filter(match_pins, net_list)],
        )
        return net

    def make_component(
        self, comp_addr: AddrStr, libsource: KicadLibpart
    ) -> KicadComponent:
        """Make a KiCAD component object from a representative instance object."""

        # TODO: improve this
        sheetpath = (
            KicadSheetpath(  # That's not actually what we want. Will have to fix
                names=comp_addr,  # TODO: going to have to strip the comp name from this
                tstamps=generate_uid_from_path(str(comp_addr)),
            )
        )

        designator = components.get_designator(comp_addr)
        constructed_component = KicadComponent(
            ref=designator,
            value=_get_value(comp_addr),
            footprint=components.get_footprint(comp_addr),
            libsource=libsource,
            tstamp=generate_uid_from_path(str(comp_addr)),
            fields=[],
            sheetpath=sheetpath,
            src_path=comp_addr,
        )
        return constructed_component

    def build(self, root) -> KicadNetlist:
        """Build a netlist from an instance"""
        self.netlist = KicadNetlist()

        # We need to cast this to a list, otherwise we're trying to reuse a generator
        all_components = list(filter(match_components, all_descendants(root)))

        # first check that all the components have a footprint
        # otherwise we can't continue the netlist build
        for cltr, component in errors.iter_through_errors(all_components):
            with cltr():
                components.get_footprint(component)

        # group the components by their footprint - because that seems
        # to be the only distinguishing feature KiCAD cares about
        for footprint, group_components in groupby(
            components.get_footprint, all_components
        ).items():
            libsource = self._libparts[footprint] = self.make_libpart(
                group_components[0]
            )

            for component in group_components:
                self._components[component] = self.make_component(component, libsource)

        for code, (net_name, pin_signal_list) in enumerate(
            nets.get_nets_by_name(root).items(), start=1
        ):
            self._nets[net_name] = self.make_net(code, net_name, pin_signal_list)

        self.netlist.libparts = list(self._libparts.values())
        self.netlist.components = list(self._components.values())
        self.netlist.nets = list(self._nets.values())

        return self.netlist


def get_netlist_as_str(root: AddrStr) -> str:
    """Return the netlist as a string."""
    builder = NetlistBuilder()
    netlist = builder.build(root)

    env = Environment(
        loader=FileSystemLoader(Path(__file__).parent), undefined=StrictUndefined
    )

    # Create the complete netlist
    template = env.get_template("kicad6.j2")
    netlist_str = template.render(nl=netlist)
    return netlist_str



from atopile.datatypes import Ref, KeyOptMap, KeyOptItem


def test_ref_from_one():
    assert Ref.from_one("foo") == ("foo",)
    assert Ref.from_one(42) == (42,)


def test_ref_add_name():
    assert Ref.from_one("foo").add_name("bar") == ("foo", "bar")


def test_keyoptitem_from_kv():
    assert KeyOptItem.from_kv(None, "foo") == (None, "foo")


def test_keyoptmap_from_item():
    assert KeyOptMap.from_item(KeyOptItem.from_kv(None, "foo")) == ((None, "foo"),)


def test_keyoptmap_from_kv():
    assert KeyOptMap.from_kv(None, "foo") == ((None, "foo"),)


def test_keyoptitem_ref():
    assert KeyOptItem((None, "foo")).ref is None
    assert KeyOptItem((Ref.from_one("foo"), "bar")).ref == ("foo",)


def test_keyoptmap_get_named_items():
    assert dict(
        KeyOptMap(
            (
                KeyOptItem((None, "foo")),
                KeyOptItem((Ref.from_one("bar"), "baz")),
            )
        ).named_items()
    ) == {
        ("bar",): "baz",
    }


def test_filter_items_by_type():
    strs = KeyOptMap(
        (
            KeyOptItem((None, "foo")),
            KeyOptItem((Ref.from_one("bar"), "baz")),
        )
    )

    ints = KeyOptMap(
        (
            KeyOptItem((None, 42)),
            KeyOptItem(("test", 43)),
        )
    )

    nones = KeyOptMap((KeyOptItem((None, None)),))

    items = KeyOptMap(strs + ints + nones)

    assert tuple(items.filter_items_by_type(int)) == ints
    assert tuple(items.filter_items_by_type(str)) == strs
    assert tuple(items.filter_items_by_type((int, str))) == strs + ints


def test_keyoptmap_get_items_by_type():
    strs = KeyOptMap(
        (
            KeyOptItem((None, "foo")),
            KeyOptItem((Ref.from_one("bar"), "baz")),
        )
    )

    ints = KeyOptMap(
        (
            KeyOptItem((None, 42)),
            KeyOptItem(("test", 43)),
        )
    )

    nones = KeyOptMap((KeyOptItem((None, None)),))

    items = KeyOptMap(strs + ints + nones)

    items_by_type = items.map_items_by_type((int, str))

    assert items_by_type[str] == strs
    assert items_by_type[int] == ints


def test_keyoptmap_get_unnamed_items():
    assert tuple(
        KeyOptMap(
            (
                KeyOptItem((None, "foo")),
                KeyOptItem((Ref.from_one("bar"), "baz")),
                KeyOptItem((None, 42)),
            )
        ).unnamed_items()
    ) == ("foo", 42)


def test_keyoptmap_keys():
    assert tuple(
        KeyOptMap(
            (
                KeyOptItem((None, "foo")),
                KeyOptItem((Ref.from_one("bar"), "baz")),
                KeyOptItem((None, 42)),
            )
        ).keys()
    ) == (("bar",),)


def test_keyoptmap_values():
    assert tuple(
        KeyOptMap(
            (
                KeyOptItem((None, "foo")),
                KeyOptItem((Ref.from_one("bar"), "baz")),
                KeyOptItem((None, 42)),
            )
        ).values()
    ) == ("foo", "baz", 42)



from typing import List, Optional

from attrs import define, field


@define
class KicadField:
    """KV pair"""

    name: str  # eg value
    value: str  # eg 10 Ohms


@define
class KicadPin:
    """
    eg. (pin (num "1") (name "") (type "passive"))
    """

    name: str
    type: str

    @property
    def num(self) -> str:
        return self.name


@define
class KicadLibpart:
    """
    eg.
    (libpart (lib "Device") (part "R")
      (description "Resistor")
      (docs "~")
      (footprints
        (fp "R_*"))
      (fields
        (field (name "Reference") "R")
        (field (name "Value") "R")
        (field (name "Datasheet") "~"))
      (pins
        (pin (num "1") (name "") (type "passive"))
        (pin (num "2") (name "") (type "passive"))))
    """

    lib: str
    part: str
    description: str
    docs: str
    footprints: List[str] = field(factory=list)
    fields: List[KicadField] = field(factory=list)
    pins: List[KicadPin] = field(factory=list)


@define
class KicadSheetpath:
    """
    eg. (sheetpath (names "/") (tstamps "/"))
    """

    names: str = "/"  # module path, eg. toy.ato/Vdiv1
    tstamps: str = "/"  # module UID, eg. b1d41e3b-ef4b-4472-9aa4-7860376ef0ce


@define
class KicadComponent:
    """
    eg.
    (comp (ref "R4")
      (value "5R")
      (libsource (lib "Device") (part "R") (description "Resistor"))
      (property (name "Sheetname") (value ""))
      (property (name "Sheetfile") (value "example.kicad_sch"))
      (sheetpath (names "/") (tstamps "/"))
      (tstamps "9c26a741-12df-4e56-baa6-794e6b3aa7cd")))
    """

    ref: str  # eg. "R1" -- should be unique, we should assign these here I think
    value: str  # eg. "10k" -- seems to be an arbitary string
    libsource: KicadLibpart
    tstamp: str  # component UID, eg. b1d41e3b-ef4b-4472-9aa4-7860376ef0ce
    src_path: str
    footprint: Optional[str] = None  # eg. "Resistor_SMD:R_0603_1608Metric"
    properties: List[KicadField] = field(factory=list)
    fields: List[KicadField] = field(factory=list)
    sheetpath: KicadSheetpath = field(factory=KicadSheetpath)


@define
class KicadNode:
    """
    eg. (node (ref "R1") (pin "1") (pintype "passive"))
    """

    ref: str
    pin: str
    pintype: str = "stereo"


@define
class KicadNet:
    """
    eg.
    (net (code "1") (name "Net-(R1-Pad1)")
      (node (ref "R1") (pin "1") (pintype "passive"))
      (node (ref "R2") (pin "1") (pintype "passive")))
    """

    code: str  # TODO: do these have to be numbers, or can they be more UIDs?
    name: str
    nodes: List[KicadNode] = field(factory=list)


class KicadLibraries:
    """
    eg.
    (libraries
    (library (logical "Device")
      (uri "/Applications/KiCad/KiCad.app/Contents/SharedSupport/symbols//Device.kicad_sym")))
    """

    def __init__(self):
        # don't believe these are mandatory and I don't think they're useful in the context of atopile
        raise NotImplementedError


@define
class KicadNetlist:
    version: str = "E"  # What does this mean?
    source: str = "unknown"  # eg. "/Users/mattwildoer/Projects/SizzleSack/schematic/sandbox.py" TODO: point this at the sourcefile
    date: str = ""  # NOTE: we don't want a data in here because it'll make a diff in our outputs -- although perhaps it's a useful field for a hash here
    tool: str = "atopile"  # TODO: add version in here too

    components: List[KicadComponent] = field(factory=list)
    libparts: List[KicadLibpart] = field(factory=list)
    nets: List[KicadNet] = field(factory=list)
    """
    (net (code "3") (name "Net-(R2-Pad2)")
      (node (ref "R2") (pin "2") (pintype "passive"))
      (node (ref "R3") (pin "1") (pintype "passive"))
      (node (ref "R4") (pin "1") (pintype "passive")))
    """



from collections import defaultdict
from typing import Iterable, Optional
from toolz import groupby
from attr import define

from atopile import address
from atopile.address import AddrStr
from atopile.datatypes import Ref
from atopile.instance_methods import (
    all_descendants,
    get_children,
    get_links,
    get_parent,
    iter_parents,
    match_interfaces,
    match_pins_and_signals,
    match_signals,
    match_modules,
)
from atopile.loop_soup import LoopSoup
from atopile.address import get_name, add_instance


def get_nets(root: AddrStr) -> Iterable[Iterable[str]]:
    """Find all the nets under a given root."""
    net_soup = LoopSoup()
    for addr in all_descendants(root):
        if match_pins_and_signals(addr):
            net_soup.add(addr)
        for source, target in get_links(addr):
            if match_interfaces(source) and match_interfaces(target):
                for int_pin in get_children(source):
                    if match_pins_and_signals(int_pin):
                        net_soup.join(int_pin, add_instance(target, get_name(int_pin)))
                    else:
                        raise NotImplementedError
            elif match_interfaces(source) or match_interfaces(source):
                raise NotImplementedError
            else:
                net_soup.join(source, target)
    return net_soup.groups()


@define
class _Net:
    nodes_on_net: list[str]

    base_name: Optional[str] = None

    lcp: Optional[str] = None
    prefix: Optional[Ref] = None

    suffix: Optional[int] = None

    def get_name(self) -> str:
        """Get the name of the net."""
        # suffix is a ref (tuple of strings)
        return (
            f"{'-'.join(self.prefix) + '-' if self.prefix else ''}"
            f"{self.base_name or 'net'}"
            f"{'-' + str(self.suffix) if self.suffix else ''}"
        )

    def generate_base_net_name(self) -> Optional[str]:
        """TODO:"""
        WEIGHT_NO_GRANDPARENTS = 10
        WEIGHT_INTERFACE_GRANDPARENT = 5
        WEIGHT_SIGNAL = 2

        name_candidates = defaultdict(int)

        for signal in filter(match_signals, self.nodes_on_net):
            name = get_name(signal)
            if get_parent(signal) is None:
                name_candidates[name] += WEIGHT_NO_GRANDPARENTS
            elif any(map(match_interfaces, iter_parents(signal))):
                name_candidates[name] += WEIGHT_INTERFACE_GRANDPARENT
            else:
                name_candidates[name] += WEIGHT_SIGNAL

        if name_candidates:
            highest_rated_name = max(name_candidates, key=name_candidates.get)
            self.base_name = highest_rated_name


def _find_net_names(nets: Iterable[Iterable[str]]) -> dict[str, list[str]]:
    """Find the names of the nets."""
    # make net objects
    net_objs = [_Net(list(net)) for net in nets]

    # grab all the nets base names
    for net in net_objs:
        net.generate_base_net_name()

    # for the net objects that still conflict, grab a prefix
    conflicing_nets = _find_conflicts(net_objs)
    _add_prefix(conflicing_nets)

    # if they still conflict, slap a suffix on that bad boi
    conflicing_nets = _find_conflicts(net_objs)
    _add_suffix(conflicing_nets)

    return {net.get_name(): net.nodes_on_net for net in net_objs}


def _find_conflicts(nets: Iterable[_Net]) -> Iterable[Iterable[_Net]]:
    """"""
    nets_grouped_by_name = groupby(lambda net: net.get_name(), nets)
    for nets in nets_grouped_by_name.values():
        if len(nets) > 1:
            yield nets


def _add_prefix(conflicts: Iterable[list[_Net]]):
    """Resolve conflicts in net names."""
    for conflict_nets in conflicts:
        for net in conflict_nets:
            if net.base_name:
                # Find the parent of the net that is a module
                parent_module_iter = filter(
                    match_modules, iter_parents(net.nodes_on_net[0])
                )

                # Get the first parent module that matches, or None if there's no match
                parent_module = next(parent_module_iter, None)

                # Check if a parent module was found
                if parent_module:
                    # Get the ref of the parent module
                    if hasattr(parent_module, "ref"):
                        net.prefix = address.get_instance_section(parent_module)


def _add_suffix(conflicts: Iterable[list[_Net]]):
    """Add an integer suffix to the nets to resolve conflicts."""
    for conflict_nets in conflicts:
        for i, net in enumerate(conflict_nets):
            net.suffix = i


class NetFinder:
    def __init__(self) -> None:
        self.net_name_to_nodes_map: dict[AddrStr, dict[str, Iterable[AddrStr]]] = {}
        self.node_to_net_name: dict[AddrStr, dict[AddrStr, str]] = {}

    def get_nets_by_name(self, entry: AddrStr) -> dict[str, list[AddrStr]]:
        """Get the nets for a given root."""
        if address.get_instance_section(entry):
            raise ValueError("Only entry are supported for now")

        if entry not in self.net_name_to_nodes_map:
            self.net_name_to_nodes_map[entry] = _find_net_names(get_nets(entry))
            self.node_to_net_name[entry] = {
                node: net_name
                for net_name, nodes in self.net_name_to_nodes_map[entry].items()
                for node in nodes
            }

        return self.net_name_to_nodes_map[entry]

    def get_net_name_node_is_on(self, node: AddrStr) -> str:
        """Get the net name for a given node."""
        entry = address.get_entry(node)
        if entry not in self.node_to_net_name:
            self.get_nets_by_name(entry)

        return self.node_to_net_name[entry][node]


net_finder = NetFinder()


def get_net_name_node_is_on(addr: str) -> str:
    """
    Return the net name for the given address
    """
    return net_finder.get_net_name_node_is_on(addr)


def get_nets_by_name(addr: AddrStr) -> dict[str, list[AddrStr]]:
    """
    Return a dict of net names to nodes
    """
    return net_finder.get_nets_by_name(addr)



# pylint: disable=logging-fstring-interpolation

"""
This CLI command provides the `ato install` command to:
- install dependencies
- download JLCPCB footprints
"""

import logging
import subprocess
from itertools import chain
from pathlib import Path
from typing import Optional

import click
import yaml
from git import InvalidGitRepositoryError, NoSuchPathError, Repo

from atopile import config, errors, version

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


@click.command("install")
@click.argument("to_install", required=False)
@click.option("--jlcpcb", is_flag=True, help="JLCPCB component ID")
@click.option("--upgrade", is_flag=True, help="Upgrade dependencies")
def install(to_install: str, jlcpcb: bool, upgrade: bool):
    """
    Install a dependency of for the project.
    """
    repo = Repo(".", search_parent_directories=True)
    top_level_path = Path(repo.working_tree_dir)

    if to_install:
        # eg. "ato install some-atopile-module"
        installed_semver = install_dependency(
            to_install,
            top_level_path,
            upgrade
        )
        module_name, module_spec = split_module_spec(to_install)
        if module_spec is None and installed_semver:
            # If the user didn't specify a version, we'll
            # use the one we just installed as a basis
            to_install = f"{module_name}^{installed_semver}"
        set_dependency_to_ato_yaml(top_level_path, to_install)

    elif jlcpcb:
        # eg. "ato install --jlcpcb=C123"
        install_jlcpcb(to_install)
    else:
        # eg. "ato install"
        install_dependencies_from_yaml(top_level_path, upgrade)

    log.info("[green]Done![/] :call_me_hand:", extra={"markup": True})


def split_module_spec(spec: str) -> tuple[str, Optional[str]]:
    """Splits a module spec string into the module name and the version spec."""
    for splitter in chain(version.OPERATORS, (" ", "@")):
        if splitter in spec:
            splitter_loc = spec.find(splitter)
            if splitter_loc < 0:
                continue

            module_name = spec[:splitter_loc].strip()
            version_spec = spec[splitter_loc:].strip()
            return module_name, version_spec

    return spec, None


def set_dependency_to_ato_yaml(top_level_path: Path, module_spec: str):
    """Add deps to the ato.yaml file"""
    # Get the existing config data
    ato_yaml_path = top_level_path / "ato.yaml"
    if not ato_yaml_path.exists():
        raise errors.AtoError(f"ato.yaml not found in {top_level_path}")

    with ato_yaml_path.open("r") as file:
        data = yaml.safe_load(file) or {}

    # Add module to dependencies, avoiding duplicates
    dependencies: list[str] = data.setdefault("dependencies", [])
    dependencies_by_name: dict[str, str] = {
        split_module_spec(x)[0]: x for x in dependencies
    }

    module_to_install, _ = split_module_spec(module_spec)
    if module_to_install in dependencies_by_name:
        existing_spec = dependencies_by_name[module_to_install]
        if existing_spec != module_spec:
            dependencies.remove(existing_spec)

    if module_spec not in dependencies:
        dependencies.append(module_spec)

        with ato_yaml_path.open("w") as file:
            yaml.safe_dump(data, file, default_flow_style=False)


def install_dependency(
    module: str, top_level_path: Path, upgrade: bool = False
) -> Optional[version.Version]:
    """
    Install a dependency of the name "module_name"
    into the project to "top_level_path"
    """
    # Figure out what we're trying to install here
    module_name, module_spec = split_module_spec(module)
    if not module_spec:
        module_spec = "*"

    # Ensure the modules path exists
    modules_path = config.get_project_config_from_path(
        top_level_path
    ).paths.abs_module_path
    modules_path.mkdir(parents=True, exist_ok=True)

    try:
        # This will raise an exception if the directory does not exist
        repo = Repo(modules_path / module_name)
    except (InvalidGitRepositoryError, NoSuchPathError):
        # Directory does not contain a valid repo, clone into it
        log.info(f"Installing dependency {module_name}")
        clone_url = f"https://gitlab.atopile.io/packages/{module_name}"
        repo = Repo.clone_from(clone_url, modules_path / module_name)
    else:
        # In this case the directory exists and contains a valid repo
        if upgrade:
            log.info(f"Fetching latest changes for {module_name}")
            repo.remotes.origin.fetch()
        else:
            log.info(
                f"{module_name} already exists. If you wish to upgrade, use --upgrade"
            )
            # here we're done because we don't want to play with peoples' deps under them
            return

    # Figure out what version of this thing we need
    semver_to_tag = {}
    installed_semver = None
    for tag in repo.tags:
        try:
            semver_to_tag[version.parse(tag.name)] = tag
        except errors.AtoError:
            log.debug(f"Tag {tag.name} is not a valid semver tag. Skipping.")

    if "@" in module_spec:
        # If there's an @ in the version, we're gonna check that thing out
        best_checkout = module_spec.strip().strip("@")
    elif semver_to_tag:
        # Otherwise we're gonna find the best tag meeting the semver spec
        valid_versions = [v for v in semver_to_tag if version.match(module_spec, v)]
        installed_semver = max(valid_versions)
        best_checkout = semver_to_tag[installed_semver]
    else:
        log.warning(
            "No semver tags found for this module. Using latest default branch :hot_pepper:.",
            extra={"markup": True},
        )

    # If the repo is dirty, throw an error
    if repo.is_dirty():
        raise errors.AtoError(
            f"Module {module_name} has uncommitted changes. Aborting."
        )

    # Checkout the best thing we've found
    commit_before_checkout = repo.head.commit
    repo.git.checkout(best_checkout)
    if repo.head.commit == commit_before_checkout:
        log.info(
            f"Already on the best option ([cyan bold]{best_checkout}[/]) for {module_name}",
            extra={"markup": True},
        )
    else:
        log.info(
            f"Using :sparkles: [cyan bold]{best_checkout}[/] :sparkles: of {module_name}",
            extra={"markup": True},
        )

    if installed_semver:
        return best_checkout


def install_dependencies_from_yaml(top_level_path: Path, upgrade: bool = False):
    """Install all dependencies from the ato.yaml file"""
    cfg = config.get_project_config_from_path(top_level_path)
    for module_name in cfg.dependencies:
        install_dependency(module_name, top_level_path, upgrade)


def install_jlcpcb(component_id: str):
    """Install a component from JLCPCB"""
    component_id = component_id.upper()
    if not component_id.startswith("C") or not component_id[1:].isdigit():
        raise errors.AtoError(f"Component id {component_id} is invalid. Aborting.")

    # Get the top level of the git module the user is currently within
    repo = Repo(".", search_parent_directories=True)
    top_level_path = Path(repo.working_tree_dir)

    # Get the remote URL
    remote_url = repo.remote().url

    # Set the footprints_dir based on the remote URL
    if remote_url == "git@gitlab.atopile.io:atopile/modules.git":
        footprints_dir = top_level_path / "footprints"
    else:
        footprints_dir = top_level_path / "elec/footprints/footprints"

    log.info(f"Footprints directory: {footprints_dir}")

    command = [
        "easyeda2kicad",
        "--full",
        f"--lcsc_id={component_id}",
        f"--output={footprints_dir}",
        "--overwrite",
        "--ato",
        f"--ato_file_path={top_level_path / 'elec/src'}",
    ]
    result = subprocess.run(command, capture_output=True, text=True, check=False)

    # The stdout and stderr are captured due to 'capture_output=True'
    print("STDOUT:", result.stdout)
    print("STDERR:", result.stderr)

    # Check the return code to see if the command was successful
    if result.returncode == 0:
        print("Command executed successfully")
    else:
        raise errors.AtoError("Couldn't install component")



"""
This is a terminal target that will generate BoMs
"""

import csv
import itertools
import logging
from collections import OrderedDict
from io import StringIO

import natsort
import rich
from rich.style import Style
from rich.table import Table
from toolz import groupby

from atopile import address, errors, components
from atopile.instance_methods import all_descendants, match_components

log = logging.getLogger(__name__)


# These functions are used to downgrade the errors to warnings.
# Those warnings are logged and the default value is returned.
_get_mpn = errors.downgrade(
    components.get_mpn, (components.MissingData, components.NoMatchingComponent)
)

_get_footprint = errors.downgrade(
    components.get_footprint, components.MissingData, default="?"
)


def _get_value(addr: address.AddrStr) -> str:
    value = errors.downgrade(
        components.get_user_facing_value,
        (components.MissingData, components.NoMatchingComponent)
    )(addr)

    if value is not None:
        return value

    value = str(errors.downgrade(
        components.get_specd_value, components.MissingData
    )(addr))

    if value is not None:
        return value

    return "?"


light_row = Style(color="bright_black")
dark_row = Style(color="white")


def generate_designator_map(entry_addr: address.AddrStr) -> str:
    """Generate a map between the designator and the component name"""

    if address.get_instance_section(entry_addr):
        raise ValueError("Cannot generate a BoM for an instance address.")

    all_components = list(filter(match_components, all_descendants(entry_addr)))

    # Create tables to print to the terminal and to the disc
    console_table = Table(show_header=True, header_style="bold green")
    console_table.add_column("Des", justify="right")
    console_table.add_column("Name", justify="left")
    console_table.add_column("Name", justify="left")
    console_table.add_column("Des", justify="left")

    # Populate the tables
    sorted_designator_dict = {}
    sorted_comp_name_dict = {}
    for component in all_components:
        c_des = components.get_designator(component)
        c_name = address.get_instance_section(component)
        sorted_designator_dict[c_des] = c_name
        sorted_comp_name_dict[c_name] = c_des

    sorted_designator_dict = OrderedDict(
        natsort.natsorted(sorted_designator_dict.items())
    )
    sorted_comp_name_dict = OrderedDict(sorted(sorted_comp_name_dict.items()))

    for row_index, ((s_des, n_comp), (s_comp, n_des)) in enumerate(
        zip(sorted_designator_dict.items(), sorted_comp_name_dict.items())
    ):
        console_table.add_row(
            s_des, n_comp, s_comp, n_des, style=dark_row if row_index % 2 else light_row
        )

    # Print the table
    rich.print(console_table)


def generate_bom(entry_addr: address.AddrStr) -> str:
    """Generate a BoM for the and print it to a CSV."""

    if address.get_instance_section(entry_addr):
        raise ValueError("Cannot generate a BoM for an instance address.")

    all_components = list(filter(match_components, all_descendants(entry_addr)))
    bom = groupby(_get_mpn, all_components)

    # JLC format: Comment (whatever might be helpful) Designator Footprint LCSC
    COLUMNS = ["Comment", "Designator", "Footprint", "LCSC"]

    # Create tables to print to the terminal and to the disc
    console_table = Table(show_header=True, header_style="bold magenta")
    for column in COLUMNS:
        console_table.add_column(column)

    csv_table = StringIO()
    writer = csv.DictWriter(csv_table, fieldnames=COLUMNS)
    writer.writeheader()

    bom_row_nb_counter = itertools.count()

    # Help to fill both tables
    def _add_row(value, designator, footprint, mpn):
        row_nb = next(bom_row_nb_counter)
        writer.writerow(
            {
                "Comment": value,
                "Designator": designator,
                "Footprint": footprint,
                "LCSC": mpn,
            }
        )
        console_table.add_row(
            value,
            designator,
            footprint,
            mpn,
            style=dark_row if row_nb % 2 else light_row,
        )

    # Populate the tables
    for mpn, components_in_group in bom.items():
        if mpn:
            # representative component
            component = components_in_group[0]

            friendly_designators = ",".join(
                components.get_designator(component)
                for component in components_in_group
            )

            _add_row(
                _get_value(component),
                friendly_designators,
                _get_footprint(component),
                mpn,
            )
        else:
            # for components without an MPN, we add a row for each component
            # this way the user can manually add the MPN as they see fit
            for component in components_in_group:
                _add_row(
                    _get_value(component),
                    components.get_designator(component),
                    _get_footprint(component),
                    "?",
                )

    # Print the table
    rich.print(console_table)

    # Return the CSV
    return csv_table.getvalue()



"""
A loop provides us a means to efficiently find connectedness between objects,
while only having access to a series of connections.
"""

import collections.abc
from typing import Callable, Hashable, Iterator, Optional, TypeVar

T = TypeVar("T")


class LoopItem(collections.abc.Iterable[T]):
    """
    Express a loop of objects, where each represents something and just points to the next
    """

    def __init__(self, represents: T):
        self.represents = represents
        self.prev = self
        self.next = self

    def iter_loop(self, limit: Optional[int] = None) -> Iterator["LoopItem[T]"]:
        """Iterate over the loop"""
        current = self
        count = 0
        yield current
        while current.next is not self:
            current = current.next
            count += 1
            if limit is not None and count >= limit:
                raise RuntimeError(f"Loop is too long, limit is {limit}")
            yield current

    def iter_values(self, limit: Optional[int] = None) -> Iterator[T]:
        """Iterate over the values in the loop"""
        for i in self.iter_loop(limit):
            yield i.represents

    def __iter__(self) -> Iterator[T]:
        yield from self.iter_values()

    def __repr__(self) -> str:
        if self.next is self:
            return f"<Loop {repr(self.represents)} -> {repr(self.represents)}>"
        return f"<Loop {repr(self.prev.represents)} -> {repr(self.represents)} -> {repr(self.next.represents)}>"

    @staticmethod
    def join(a: "LoopItem", b: "LoopItem") -> None:
        """Join two loops together"""

        # if they're the same, do nothing
        if a is b:
            return

        # if they're both lonely, make them friends
        if a.next is a and b.next is b:
            assert a.prev is a
            assert b.prev is b
            a.next = b
            a.prev = b
            b.next = a
            b.prev = a

        # if one is lonely, make it friends with the other`
        elif a.next is a:
            assert a.prev is a
            old_next = b.next
            b.next = a
            a.prev = b
            a.next = old_next
            old_next.prev = a
        elif b.next is b:
            assert b.prev is b
            old_next = a.next
            a.next = b
            b.prev = a
            b.next = old_next
            old_next.prev = b

        # if neither is lonely, check if they are already joined
        # If not, join them
        else:
            a_loop_items = list(a.iter_loop())
            if b not in a_loop_items:
                a_old_next = a.next
                b_old_prev = b.prev
                a.next = b
                b.prev = a
                b_old_prev.next = a_old_next
                a_old_next.prev = b_old_prev


def _simple_return(x: T) -> T:
    return x


class LoopSoup:
    """
    Helper function to associate data with the loop class
    """

    def __init__(self, key_func: Callable[[T], Hashable] = _simple_return):
        self.key_func = key_func
        self._map: dict[Hashable, T] = {}

    def get_loop(self, thing: T) -> LoopItem[T]:
        """Get the loop for a thing"""
        return self._map[self.key_func(thing)]

    def add(self, thing: T) -> LoopItem[T]:
        """Add a thing to the loop pool"""
        key = self.key_func(thing)
        if key in self._map:
            raise KeyError(f"Key {key} for {repr(thing)} already occupied")
        loop_item = LoopItem(thing)
        self._map[key] = loop_item
        return loop_item

    def join(self, a: T, b: T) -> None:
        """Join two things together"""
        LoopItem.join(self.get_loop(a), self.get_loop(b))

    def groups(self) -> Iterator[tuple[T]]:
        """Return an iterator of groups of things that are connected together"""
        seen = set()

        for k, v in self._map.items():
            if k in seen:
                continue

            # we can't do this bit lazily because we need to know
            # which values we would see if we went through the whole group
            values = tuple(v.iter_values())
            seen.update(self.key_func(i) for i in values)
            yield values



import pytest
from semver import Version

from atopile import errors
from atopile.version import match, parse

# disable docstring checks - it's a test file
# pylint: disable=C0116


@pytest.fixture
def version():
    return parse("1.2.3")


def test_match_equal(version):
    assert match("==1.2.3", version)
    assert match("==1.2.4", version) is False
    assert match("==1.2.4", version) is False


def test_match_greater_than(version):
    assert match(">1.2.2", version)
    assert match(">1.2.3", version) is False
    assert match(">1.2.4", version) is False


def test_match_less_than(version):
    assert match("<1.2.4", version)
    assert match("<1.2.3", version) is False
    assert match("<1.2.2", version) is False


def test_match_greater_than_or_equal(version):
    assert match(">=1.2.2", version)
    assert match(">=1.2.3", version)
    assert match(">=1.2.4", version) is False


def test_match_less_than_or_equal(version):
    assert match("<=1.2.4", version)
    assert match("<=1.2.3", version)
    assert match("<=1.2.2", version) is False


def test_match_caret(version):
    assert match("^1.2.3", version)
    assert match("^1.3.0", version) is False
    assert match("^2.0.0", version) is False


def test_match_dirty():
    version = parse("0.0.17-dev6+gdaf6b51")
    assert match("^0.0.16", version)
    assert match("^0.0.17", version)
    assert match("^0.0.18", version) is False


def test_match_tilde(version):
    assert match("~1.2.3", version)
    assert match("~1.3.0", version) is False
    assert match("~2.0.0", version) is False


def test_match_multiple(version):
    assert match(">=1.2.0 <1.3.0", version)
    assert match(">=1.2.0 <1.2.3", version) is False
    assert match(">=1.2.0 <1.2.4", version)


def test_match_union(version):
    assert match(">=1.2.0 || >=2.0.0", version)
    assert match(">=1.3.0 || >=2.0.0", version) is False
    assert match(">=2.0.0 || >=1.2.0", version)


def test_match_negation(version):
    assert match("!1.2.3", version) is False
    assert match("!1.2.4", version)


def test_syntax_error(version):
    with pytest.raises(errors.AtoError):
        match("abc", version)


def test_parse_valid_version():
    assert parse("1.2.3") == Version(1, 2, 3)


def test_parse_version_with_build_info():
    assert parse("1.2.3-a1") == Version(1, 2, 3, "a1")


def test_parse_version_with_prerelease_and_build_info():
    assert parse("1.2.3+build123") == Version(1, 2, 3, None, "build123")


def test_parse_version_with_hatch_shenanigans():
    assert parse("0.0.17.dev0+g0151069.d20230928") == Version(
        0, 0, 17, "dev0", "g0151069.d20230928"
    )


def test_parse_invalid_version():
    with pytest.raises(ValueError):
        parse("not a version")


def test_v_prefix(version):
    assert parse("v1.2.3") == Version(1, 2, 3)
    assert match("==v1.2.3", version)
    assert match("==v1.2.4", version) is False
    assert match("==v1.2.4", version) is False


def test_stringify(version):
    assert str(parse("v1.2.3")) == "1.2.3"



"""
Datatypes used in the model.
"""
import logging
from typing import Generic, Iterable, Iterator, Mapping, Optional, Type, TypeVar

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class Ref(tuple[str]):
    """Shell class to provide basic utils for a reference."""

    def add_name(self, name: str | int) -> "Ref":
        """Return a new Ref with the given name."""
        return Ref((*self, name))

    def __str__(self) -> str:
        return ".".join(map(str, self))

    @classmethod
    def empty(cls) -> "Ref":
        """Return an empty Ref."""
        return cls(())

    @classmethod
    def from_one(cls, name: str | int) -> "Ref":
        """Return a Ref with a single item."""
        return cls((name,))


T = TypeVar("T")


class KeyOptItem(tuple[Optional[Ref], T], Generic[T]):
    """A class representing anf optionally-named thing."""

    @property
    def ref(self) -> Optional[Ref]:
        """Return the name of this item, if it has one."""
        return self[0]

    @property
    def value(self) -> T:
        """Return the value of this item."""
        return self[1]

    @classmethod
    def from_kv(cls, key: Optional[Ref], value: T) -> "KeyOptItem[T]":
        """Return a KeyOptItem with a single item."""
        return KeyOptItem((key, value))


class KeyOptMap(tuple[KeyOptItem[T]]):
    """A class representing a set of optionally-named things."""

    def named_items(self) -> Mapping[Ref, T]:
        """Return all the named items in this set, ignoring the unnamed ones."""
        return dict(filter(lambda x: x.ref is not None, self))

    def map_items_by_type(
        self, types: Iterable[Type | Iterable[Type]]
    ) -> Mapping[Type, "KeyOptMap[T]"]:
        """Return a mapping of items in this set by type."""
        return {t: tuple(self.filter_items_by_type(t)) for t in types}

    def unnamed_items(self) -> Iterable[T]:
        """Return an interable of all the unnamed items in this set."""
        return map(lambda x: x.value, filter(lambda x: x.ref is None, self))

    def filter_items_by_type(self, types: Type | Iterable[Type]) -> Iterator[T]:
        """Helper function to filter by type."""
        return filter(lambda x: isinstance(x.value, types), self)

    def keys(self) -> Iterable[Ref]:
        """Return an iterable of all the names in this set."""
        return map(lambda x: x.ref, filter(lambda x: x.ref is not None, self))

    def values(self) -> Iterable[T]:
        """Return an iterable of all the values in this set."""
        return map(lambda x: x.value, self)

    @classmethod
    def from_item(cls, item: KeyOptItem[T]) -> "KeyOptMap[T]":
        """Return a KeyOptMap with a single item."""
        return KeyOptMap((item,))

    @classmethod
    def from_kv(cls, key: Optional[Ref], value: T) -> "KeyOptMap[T]":
        """Return a KeyOptMap with a single item."""
        return cls.from_item(KeyOptItem.from_kv(key, value))

    @classmethod
    def empty(cls) -> "KeyOptMap[T]":
        """Return an empty KeyOptMap."""
        return cls(())



import logging
from contextlib import contextmanager
from pathlib import Path

from antlr4 import CommonTokenStream, FileStream, InputStream
from antlr4.error.ErrorListener import ErrorListener

from atopile.parser.AtopileLexer import AtopileLexer
from atopile.parser.AtopileParser import AtopileParser

from .errors import AtoSyntaxError, AtoFileNotFoundError

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class ErrorListenerConverter(ErrorListener):
    """Converts an error into an AtoSyntaxError."""

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e: Exception):
        raise AtoSyntaxError.from_token(offendingSymbol, f"{str(e)} '{msg}'", )


class ErrorListenerCollector(ErrorListenerConverter):
    """Collects errors into a list."""

    def __init__(self) -> None:
        self.errors = []
        super().__init__()

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e: Exception):
        self.errors.append(
            AtoSyntaxError.from_token(offendingSymbol, f"{str(e)} '{msg}'")
        )


def make_parser(src_stream: InputStream) -> AtopileParser:
    """Make a parser from a stream."""
    lexer = AtopileLexer(src_stream)
    stream = CommonTokenStream(lexer)
    parser = AtopileParser(stream)

    return parser


def set_error_listener(parser: AtopileParser, error_listener: ErrorListener) -> None:
    """Utility function to set the error listener on a parser."""
    parser.removeErrorListeners()
    parser.addErrorListener(error_listener)


@contextmanager
def defer_parser_errors(parser: AtopileParser) -> None:
    """Defer errors from a parser until the end of the context manager."""
    error_listener = ErrorListenerCollector()
    set_error_listener(parser, error_listener)

    yield parser

    if error_listener.errors:
        raise ExceptionGroup("Errors caused parsing failure", error_listener.errors)


def parse_text_as_file(
    src_code: str, src_path: None | str | Path = None
) -> AtopileParser.File_inputContext:
    """Parse a string as a file input."""
    input = InputStream(src_code)
    input.name = src_path
    parser = make_parser(input)
    with defer_parser_errors(parser):
        tree = parser.file_input()

    return tree


def parse_file(src_path: Path) -> AtopileParser.File_inputContext:
    """Parse a file from a path."""
    input = FileStream(str(src_path), encoding="utf-8")
    input.name = src_path
    parser = make_parser(input)
    with defer_parser_errors(parser):
        tree = parser.file_input()

    return tree


class FileParser:
    """Parses a file."""

    def __init__(self) -> None:
        self.cache = {}

    def get_ast_from_file(
        self, src_path: str | Path
    ) -> AtopileParser.File_inputContext:
        """Get the AST from a file."""
        if src_path not in self.cache:
            src_path = Path(src_path)
            if not src_path.exists():
                raise AtoFileNotFoundError(src_path)
            self.cache[src_path] = parse_file(src_path)

        return self.cache[src_path]


parser = FileParser()
